(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * L1 definitions and basic lemmas.
 *)

theory L1Defs
imports CCorresE AutoCorresAttributes
begin

(* Definitions of constants used in the SimplConv output. *)
type_synonym 's L1_monad = "('s, unit + unit) nondet_monad"
definition "L1_seq (A :: 's L1_monad) (B :: 's L1_monad) \<equiv> (A >>=E (\<lambda>_. B)) :: 's L1_monad"
definition "L1_skip \<equiv> returnOk () :: 's L1_monad"
definition "L1_modify m \<equiv> liftE (modify m) :: 's L1_monad"
definition "L1_condition c (A :: 's L1_monad) (B :: 's L1_monad) \<equiv> condition c A B"
definition "L1_catch (A :: 's L1_monad) (B :: 's L1_monad) \<equiv> (A <handle> (\<lambda>_. B))"
definition "L1_while c (A :: 's L1_monad) \<equiv> (whileLoopE (\<lambda>_. c) (\<lambda>_. A) ())"
definition "L1_throw \<equiv> throwError () :: 's L1_monad"
definition "L1_spec r \<equiv> liftE (spec r) :: 's L1_monad"
definition "L1_guard c \<equiv> (liftE (guard c)) :: 's L1_monad"
definition "L1_init v \<equiv> (doE x \<leftarrow> liftE (select UNIV); liftE (modify (v (\<lambda>_. x))) odE) :: 's L1_monad"
definition "L1_call (measure_val :: 's \<Rightarrow> nat) scope_setup (dest_fn :: nat \<Rightarrow> 's L1_monad) scope_teardown return_xf \<equiv>
      doE
        s \<leftarrow> liftE get;
        ((doE liftE (modify scope_setup);
              ss \<leftarrow> liftE get;
              dest_fn (measure_val ss) odE)
         <handle> (\<lambda>_. fail));
        t \<leftarrow> liftE get;
        liftE (modify (scope_teardown s));
        liftE (modify (return_xf t))
     odE"
definition "L1_fail \<equiv> fail :: 's L1_monad"
definition "L1_recguard n b \<equiv> (condition (\<lambda>s. n = (0 :: nat)) fail b) :: 's L1_monad"
definition "L1_set_to_pred S \<equiv> \<lambda>s. s \<in> S"
definition "recguard_dec (n :: nat) \<equiv> n - 1"

lemmas L1_defs = L1_seq_def L1_skip_def L1_modify_def L1_condition_def
    L1_catch_def L1_while_def L1_throw_def L1_spec_def L1_guard_def
    L1_fail_def L1_init_def L1_recguard_def L1_set_to_pred_def

declare L1_set_to_pred_def [simp]

(* Our corres rule converting from the deeply-embedded SIMPL to a shallow embedding. *)
definition
  L1corres :: "('p \<Rightarrow> ('s, 'p, 'e) com option)
                     \<Rightarrow> ('s, unit + unit) nondet_monad \<Rightarrow> ('s, 'p, 'e) com \<Rightarrow> bool"
where
  "L1corres \<Gamma> \<equiv>
   \<lambda>A C. \<forall>s. \<not> snd (A s) \<longrightarrow>
       ((\<forall>t. \<Gamma> \<turnstile> \<langle>C, Normal s\<rangle> \<Rightarrow> t \<longrightarrow>
        (case t of
              Normal s' \<Rightarrow> (Inr (), s') \<in> fst (A s)
            | Abrupt s' \<Rightarrow> (Inl (), s') \<in> fst (A s)
            | _ \<Rightarrow> False))
        \<and> (\<Gamma> \<turnstile> C \<down> Normal s))"

lemma L1corres_alt_def: "L1corres \<Gamma> = ccorresE (\<lambda>x. x) \<Gamma> \<top> UNIV"
  apply (rule ext)+
  apply (clarsimp simp: L1corres_def ccorresE_def)
  done

lemma L1corres_skip:
  "L1corres \<Gamma> L1_skip SKIP"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: liftE_def return_def L1_skip_def returnOk_def)
  apply (clarsimp simp: ccorresE_def)
  apply (rule conjI)
   apply clarsimp
   apply (erule exec_Normal_elim_cases)
   apply (clarsimp simp: split_def return_def bindE_def bind_def)
  apply (rule terminates.Skip)
  done

lemma L1corres_throw:
  "L1corres \<Gamma> L1_throw Throw"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: ccorresE_def L1_throw_def)
  apply (rule conjI)
   apply clarsimp
   apply (erule exec_Normal_elim_cases)
   apply (clarsimp simp: throwError_def return_def)
  apply (rule terminates.Throw)
  done

lemma L1corres_seq:
  "\<lbrakk> L1corres \<Gamma> L L'; L1corres \<Gamma> R R' \<rbrakk> \<Longrightarrow>
    L1corres \<Gamma> (L1_seq L R) (L' ;; R')"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: L1_seq_def)
  apply (rule ccorresE_Seq)
  apply auto
  done

lemma L1corres_modify:
  "L1corres \<Gamma> (L1_modify m) (Basic m)"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: liftE_def modify_def get_def put_def bind_def L1_modify_def)
  apply (clarsimp simp: ccorresE_def)
  apply (auto elim: exec_Normal_elim_cases simp: split_def return_def intro: terminates.Basic)
  done

lemma L1corres_condition:
  "\<lbrakk> L1corres \<Gamma> L L'; L1corres \<Gamma> R R' \<rbrakk> \<Longrightarrow>
    L1corres \<Gamma> (L1_condition (L1_set_to_pred c) L R) (Cond c L' R')"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: L1_defs)
  apply (rule ccorresE_Cond_match)
    apply (erule ccorresE_guard_imp, simp+)[1]
   apply (erule ccorresE_guard_imp, simp+)[1]
  apply simp
  done

lemma L1corres_catch:
  "\<lbrakk> L1corres \<Gamma> L L'; L1corres \<Gamma> R R' \<rbrakk> \<Longrightarrow>
    L1corres \<Gamma> (L1_catch L R) (Catch L' R')"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: L1_catch_def)
  apply (erule ccorresE_Catch)
  apply force
  done

lemma L1corres_while:
  "\<lbrakk> L1corres \<Gamma> B B' \<rbrakk> \<Longrightarrow>
      L1corres \<Gamma> (L1_while (L1_set_to_pred c) B) (While c B')"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: L1_defs)
  apply (rule ccorresE_While)
   apply clarsimp
  apply simp
  done

lemma L1corres_guard:
  "\<lbrakk> L1corres \<Gamma> B B' \<rbrakk> \<Longrightarrow>
      L1corres \<Gamma> (L1_seq (L1_guard (L1_set_to_pred c)) B) (Guard f c B')"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: ccorresE_def L1_defs)
  apply (erule_tac x=s in allE)
  apply (rule conjI)
   apply clarsimp
   apply (erule exec_Normal_elim_cases)
    apply (monad_eq simp: Bex_def split: xstate.splits)
    apply blast
   apply (monad_eq simp: Bex_def split: xstate.splits)
  apply (monad_eq simp: Bex_def split: xstate.splits)
  apply (fastforce intro!: terminates.Guard)
  done

lemma L1corres_spec:
  "L1corres \<Gamma> (L1_spec x) (Spec x)"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: ccorresE_def L1_spec_def)
  apply (rule conjI)
   apply clarsimp
   apply (erule exec_Normal_elim_cases)
    apply (clarsimp simp: liftE_def spec_def bind_def return_def)
   apply (clarsimp simp: liftE_def spec_def bind_def return_def)
  apply (rule terminates.Spec)
  done

lemma liftE_get_bindE:
  "(liftE get >>=E B) = (\<lambda>s. B s s)"
  apply (monad_eq simp: Ball_def Bex_def)
  done

lemma L1corres_call:
  "\<And>t. \<lbrakk> L1corres \<Gamma> (\<lambda>s. dest_fn (m s) s) (Call dest) \<rbrakk> \<Longrightarrow>
    L1corres \<Gamma>
        (L1_call m scope_setup dest_fn scope_teardown f)
        (call scope_setup dest scope_teardown (\<lambda>_ t. Basic (f t)))"
  apply (clarsimp simp: L1corres_alt_def)
  apply (unfold call_def block_def L1_call_def)
  apply (rule ccorresE_DynCom)
  apply clarsimp
  apply (rule ccorresE_get)
  apply clarsimp
  apply (rule ccorresE_assume_pre, clarsimp)
  apply (rule ccorresE_guard_imp_stronger)
    apply (rule ccorresE_Seq)
     apply (rule ccorresE_Catch)
      apply (rule ccorresE_Seq)
       apply (rule L1corres_modify[unfolded L1_modify_def L1corres_alt_def])
      apply (simp only: liftE_get_bindE)
     apply (rule ccorresE_fail)
    apply (rule ccorresE_DynCom)
    apply (rule ccorresE_get)
    apply (rule ccorresE_assume_pre, clarsimp)
    apply (rule ccorresE_guard_imp)
      apply (rule ccorresE_Seq)
       apply (rule L1corres_modify[unfolded L1_modify_def L1corres_alt_def])
      apply (rule L1corres_modify[unfolded L1_modify_def L1corres_alt_def])
     apply simp
    apply simp
   apply simp
  apply simp
  done

lemma L1corres_fail:
  "L1corres \<Gamma> L1_fail X"
  apply (clarsimp simp: L1corres_alt_def)
  apply (clarsimp simp: L1_fail_def)
  apply (rule ccorresE_fail)
  done

lemma L1corres_recguard:
  "L1corres \<Gamma> A B \<Longrightarrow> L1corres \<Gamma> (L1_recguard n A) B"
  apply (clarsimp simp: L1_recguard_def)
  apply (case_tac "n = 0")
   apply clarsimp
   apply (rule L1corres_fail [unfolded L1_fail_def])
  apply clarsimp
  done

lemma L1corres_recguard_0:
  "L1corres \<Gamma> (L1_recguard 0 x) y"
  apply (clarsimp simp: L1_recguard_def)
  apply (rule L1corres_fail [unfolded L1_fail_def])
  done

lemma L1_recguard_cong [fundef_cong, cong]:
  "\<lbrakk> n = n'; n \<noteq> 0 \<Longrightarrow> b = b' \<rbrakk> \<Longrightarrow> L1_recguard n b = L1_recguard n' b'"
  apply (clarsimp simp: L1_recguard_def)
  done

lemma L1corres_prepend_unknown_var':
  "\<lbrakk> L1corres \<Gamma> A C; \<And>s::'s::type. X (\<lambda>_::'a::type. (X' s)) s = s \<rbrakk> \<Longrightarrow> L1corres \<Gamma> (L1_seq (L1_init X) A) C"
  apply atomize
  apply (clarsimp simp: L1corres_alt_def)
  apply (unfold L1_seq_def)
  apply (rule ccorresE_symb_exec_l, assumption)
   apply (subst L1_init_def)
   apply (clarsimp simp: exs_valid_def)
   apply (monad_eq simp: Bex_def)
   apply metis
  apply (subst L1_init_def)
  apply (wp select_wp)
  apply fastforce
  done

lemma L1_catch_seq_join: "no_throw \<top> A \<Longrightarrow> L1_seq A (L1_catch B C) = (L1_catch (L1_seq A B) C)"
  apply (monad_eq simp: L1_seq_def L1_catch_def Bex_def
             Ball_def no_throw_def' split: sum.splits)
  apply blast
  done

lemma no_throw_L1_init [simp]: "no_throw P (L1_init f)"
  apply (unfold L1_init_def)
  apply (rule no_throw_bindE [where B=\<top>])
    apply simp
   apply simp
  apply wp
  done

lemma L1corres_prepend_unknown_var:
  "\<lbrakk> L1corres \<Gamma> (L1_catch A B) C; \<And>s. X (\<lambda>_::'d::type. (X' s)) s = s \<rbrakk>
         \<Longrightarrow> L1corres \<Gamma> (L1_catch (L1_seq (L1_init X) A) B) C"
  apply atomize
  apply (clarsimp simp: L1corres_alt_def)
  apply (unfold L1_seq_def L1_catch_def)
  apply (subst bindE_handleE_join [symmetric])
   apply simp
  apply (erule L1corres_prepend_unknown_var' [unfolded L1_seq_def L1_spec_def L1corres_alt_def])
  apply force
  done

lemma L1corres_prepend_unknown_var_recguard:
  "\<lbrakk> L1corres \<Gamma> (L1_recguard n A) C; \<And>s. X (\<lambda>_::'a::type. X' s) s = s\<rbrakk>
    \<Longrightarrow> L1corres \<Gamma> (L1_recguard n (L1_seq (L1_init X) A)) C"
  apply (clarsimp simp: L1_recguard_def)
  apply (case_tac n)
   apply clarsimp
  apply clarsimp
  apply (erule L1corres_prepend_unknown_var')
  apply assumption
  done

lemma L1corres_Call:
  "\<lbrakk> \<Gamma> X' = Some Z'; L1corres \<Gamma> Z Z' \<rbrakk> \<Longrightarrow>
    L1corres \<Gamma> Z (Call X')"
  apply (clarsimp simp: L1corres_alt_def)
  apply (erule (1) ccorresE_Call)
  done

lemma L1_call_corres [fundef_cong]:
  "\<lbrakk> measure_val = measure_val';
     scope_setup = scope_setup';
     \<And>s. dest_fn (measure_val s) = dest_fn' (measure_val' s);
     scope_teardown = scope_teardown';
     return_xf = return_xf' \<rbrakk> \<Longrightarrow>
    L1_call measure_val scope_setup dest_fn scope_teardown return_xf =
    L1_call measure_val' scope_setup' dest_fn' scope_teardown' return_xf'"
  apply (clarsimp simp: L1_call_def)
  done

(* Saying \<And>measure is the same thing as \<And>s *)
lemma L1corres_measure_from_state:
  "\<And>\<Gamma> l1_f m f. (\<And>(measure' :: nat). L1corres \<Gamma> (l1_f measure') f) \<Longrightarrow> L1corres \<Gamma> (\<lambda>s. l1_f (m s) s) f"
  apply (unfold L1corres_def)
  apply simp
  done

declare recguard_dec_def [termination_simp]

end
