(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * More theories and lemmas related to the non-deterministic monad library.
 *)

theory NonDetMonadEx
imports
  "../lib/wp/NonDetMonad"
  "../lib/NonDetMonadLemmaBucket"
  "../lib/OptionMonadND"
begin

(*
 * Definitions.
 *)

(* Fail unless the given state obeys the given condition. *)
abbreviation "guard \<equiv> state_assert"

lemma guard_def:
    "guard P = (\<lambda>s. (if P s then {((), s)} else {}, \<not> P s))"
  apply (monad_eq simp: state_assert_def Bex_def)
  done

abbreviation "spec \<equiv> state_select"
lemmas spec_def = state_select_def

(* Fail unless the given state obeys the given condition. *)
definition
  guardE :: "('s \<Rightarrow> bool) \<Rightarrow> ('s, 'e + unit) nondet_monad"
where
  "guardE P \<equiv> (liftE (guard P))"

(* Shorthand definitions. *)
definition "unknown \<equiv> select UNIV"
definition "selectE S = liftE (select S)"
definition "unknownE \<equiv> selectE UNIV"
definition "skip \<equiv> return ()"
definition "skipE \<equiv> returnOk ()"
definition "modifyE m = liftE (modify m)"
definition "specE r \<equiv> liftE (spec r)"
definition "getsE g \<equiv> liftE (gets g)"

lemmas monad_defs = unknown_def unknownE_def selectE_def skip_def
    skipE_def modifyE_def specE_def getsE_def guardE_def

(*
 * Theorems
 *)

lemma guardE_wp [wp]: "\<lbrace> \<lambda>s. f s \<longrightarrow> P () s \<rbrace> guardE f \<lbrace> P \<rbrace>,\<lbrace> Q \<rbrace>"
  apply (unfold guardE_def)
  apply wp
  done

(*
lemma handleE'_cong:
  "\<lbrakk>M = M'; \<And>v s s'. (Inl v, s') \<in> fst (M' s) \<Longrightarrow> N v s' = N' v s'\<rbrakk> \<Longrightarrow> (M <handle2> N) = (M' <handle2> N')"
  apply (simp add: handleE'_def)
  apply (rule bind_cong)
   apply (rule refl)
  apply (clarsimp split: sum.splits)
  done

lemma handleE'_apply_cong:
  "\<lbrakk>M s = M' s'; \<And>v st. (Inl v, st) \<in> fst (M' s') \<Longrightarrow> N v st = N' v st \<rbrakk> \<Longrightarrow> (M <handle2> N) s = (M' <handle2> N') s'"
  apply (simp add: handleE'_def)
  apply (erule bind_apply_cong)
  apply (clarsimp split: sum.splits)
  done

lemma handleE_cong:
  "\<lbrakk>M = M'; \<And>v s s'. (Inl v, s') \<in> fst (M' s) \<Longrightarrow> N v s' = N' v s'\<rbrakk> \<Longrightarrow> (M <handle> N) = (M' <handle> N')"
  apply (unfold handleE_def)
  apply (rule handleE'_cong)
  apply auto
  done

lemma handleE_apply_cong:
  "\<lbrakk>M s = M' s'; \<And>v st. (Inl v, st) \<in> fst (M' s') \<Longrightarrow> N v st = N' v st \<rbrakk> \<Longrightarrow> (M <handle> N) s = (M' <handle> N') s'"
  apply (unfold handleE_def)
  apply (rule handleE'_apply_cong)
  apply auto
  done

*)
(*
lemma liftE_apply_cong: "\<lbrakk> a = a'; X a' = X' a' \<rbrakk> \<Longrightarrow> liftE X a = liftE X' a'"
  apply (clarsimp simp: liftE_def return_def bind_def)
  done

*)

lemma empty_fail_skip [simp]: "empty_fail skip"
  apply (clarsimp simp: empty_fail_def skip_def return_def)
  done

lemma empty_fail_exception [simp]:
  "empty_fail (modifyE m)"
  "empty_fail skipE"
  "empty_fail (guardE G)"
  "empty_fail (returnOk r)"
  "empty_fail (getsE g)"
  apply (auto simp add: skipE_def modifyE_def returnOk_def guardE_def getsE_def)
  done

lemma simple_nothrow [simp]:
  "no_throw \<top> skipE"
  "no_throw \<top> (modifyE m)"
  "no_throw \<top> (guardE g)"
  "no_throw \<top> (specE a)"
  "no_throw \<top> (getsE v)"
  "no_throw \<top> fail"
  apply (auto simp: skipE_def modifyE_def guardE_def specE_def getsE_def)
  done

lemma simple_bindE_fail [simp]:
  "(guardE X >>=E (\<lambda>_. fail)) = fail"
  "(modifyE M >>=E (\<lambda>_. fail)) = fail"
  "(returnOk X >>=E (\<lambda>_. fail)) = fail"
  "(getsE X >>=E (\<lambda>_. fail)) = fail"
  "(skipE >>=E (\<lambda>_. fail)) = fail"
  "handleE fail x = fail"
  apply (auto intro!: bindE_fail_propagates simp: handleE_def handleE'_def)
  done

lemma skip_wp [wp]: "\<lbrace> P () \<rbrace> skip \<lbrace> P \<rbrace>"
  apply (clarsimp simp: skip_def valid_def return_def)
  done

lemma skip_nf [wp]: "\<lbrace> P () \<rbrace> skip \<lbrace> P \<rbrace>!"
  apply (monad_eq simp: validNF_def valid_def no_fail_def skip_def)
  done

declare valid_prod_case [wp]
declare validE_prod_case [wp]

lemma getsE_to_returnOk [simp]: "getsE (\<lambda>s. v) = returnOk v"
  apply (monad_eq simp: getsE_def)
  done

lemma guard_conj_split:
  "guard (\<lambda>s. P s \<and> Q s) = (guard P >>= (\<lambda>_. guard Q))"
  apply (monad_eq simp: Bex_def)
  done

lemma validNF_assume_pre:
  "\<lbrakk> \<And>s. P s \<Longrightarrow> \<lbrace>P\<rbrace> f \<lbrace>Q\<rbrace>! \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> f \<lbrace>Q\<rbrace>!"
    by (metis hoare_assume_pre no_fail_assume_pre validNF_def)

lemma liftE_condition_distrib: "liftE (condition C L R) = condition C (liftE L) (liftE R)"
  apply monad_eq
  apply force
  done

lemma condition_gets:
  "condition C (gets P) (gets Q) = gets (\<lambda>s. if C s then P s else Q s)"
  apply (monad_eq)
  done

lemma if_distrib_conj: "(if P then (A \<and> B) else (A \<and> C)) = (A \<and> (if P then B else C))"
  apply clarsimp
  done

lemma if_distrib_disj: "(if P then (A \<or> B) else (A \<or> C)) = (A \<or> (if P then B else C))"
  apply clarsimp
  done

lemma if_distrib_disj': "(if P then (B \<or> A) else (C \<or> A)) = (A \<or> (if P then B else C))"
  apply clarsimp
  apply blast
  done

lemma whileLoop_fail:
    "(whileLoop C (\<lambda>_. fail) i) = (guard (\<lambda>s. \<not> C i s) >>= (\<lambda>_. return i))"
  apply (subst whileLoop_unroll)
  apply (monad_eq)
  done

lemma whileLoopE_fail:
    "(whileLoopE C (\<lambda>_. fail) i) = (guardE (\<lambda>s. \<not> C i s) >>= (\<lambda>_. returnOk i))"
  apply (subst whileLoopE_unroll)
  apply (monad_eq simp: guardE_def)
  done

lemma oreturn_exec [simp]:
    "oreturn a s = Some a"
  by (clarsimp simp: oreturn_def)

lemma ogets_exec [simp]:
    "ogets a s = Some (a s)"
  by (clarsimp simp: ogets_def)

lemma ofail_exec [simp]:
    "ofail s = None"
  by (clarsimp simp: ofail_def)

lemma oguard_failed [simp]:
    "(oguard G s = None) = (\<not> G s)"
  by (clarsimp simp: oguard_def)

lemma oguard_passed [simp]:
    "(oguard G s = Some v) = (G s)"
  by (clarsimp simp: oguard_def)

lemma owhile_fail:
    "(owhile C (\<lambda>_. ofail) i) = (oguard (\<lambda>s. \<not> C i s) |>> (\<lambda>_. oreturn i))"
  apply (rule ext)
  apply (subst owhile_unroll)
  apply (clarsimp simp: obind_def ocondition_def split: option.splits)
  done

lemma liftE_left: "x >>= (\<lambda>a. liftE (y a)) \<equiv> liftE (x >>= y)"
  by (simp add: bind_liftE_distrib liftE_bindE)

lemma liftE_liftE: "liftE x >>=E (\<lambda>a. liftE (y a)) \<equiv> liftE (x >>= y)"
  by (simp add: bind_liftE_distrib)

lemma liftE_gets_fail: "liftE (gets M) >>=E (\<lambda>_. fail) = fail"
  by (metis gets_bind_ign liftE_bindE)


lemma gets_the_to_return:
   "X s = Some v \<Longrightarrow> gets_the X s = return v s"
  by (clarsimp simp: gets_the_def exec_gets)

lemma owhile_to_fold:
  assumes p_increments: "\<And>r. P (Q r) = (P r + 1 :: word32)"
  shows "owhile (\<lambda>r s. P r < x) (\<lambda>r. oreturn (Q r)) i =
      oreturn (if P i \<le> x then fold (\<lambda>i. Q) [unat (P i)..<unat x] i else i)"
    (is "?LHS = oreturn (?RHS x)")
  apply (rule ext)
  apply (case_tac "P i \<le> x")
   apply simp
   apply (rule_tac s=xa in owhile_rule [where I="\<lambda>r s. P i \<le> P r \<and> P r \<le> x \<and> r = ?RHS (P r)"
        and M="measure (\<lambda>r. unat x - unat (P r))"])
        apply clarsimp
       apply clarsimp
      apply clarsimp
      apply (clarsimp simp: p_increments)
      apply (subst unatSuc2)
       apply (metis less_is_non_zero_p1)
      apply unat_arith
     apply (clarsimp simp: p_increments)
     apply (subst unat_Suc2)
      apply clarsimp
     apply clarsimp
     apply unat_arith
    apply (clarsimp simp: oreturn_def)
   apply (clarsimp simp: oreturn_def)
  apply simp
  apply (subst owhile_unroll)
  apply (subst ocondition_False)
   apply simp
  apply simp
  done

lemma whileLoop_to_fold:
  assumes p_increments: "\<And>r. P (Q r) = (P r + 1 :: word32)"
  shows "(whileLoop (\<lambda>r s. P r < x)
           (\<lambda>r. return (Q r))
           i s) = return (if P i \<le> x then fold (\<lambda>i r. (Q r)) [unat (P i) ..< unat x] i else i) s"
    (is "?LHS s = return (?RHS x) s")
  apply (subst OptionMonadND.gets_the_return [symmetric])
  apply (subst gets_the_whileLoop)
  apply (rule gets_the_to_return)
  apply (subst owhile_to_fold)
   apply (rule p_increments)
  apply (clarsimp simp: oreturn_def)
  done

end
