(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Test force/prevent heap abstraction.
 *)

theory heap_lift_force_prevent
imports "../../AutoCorres"
begin

install_C_file "heap_lift_force_prevent.c"

autocorres [
    no_heap_abs = unlifted_a unlifted_b,
    force_heap_abs = lifted_a lifted_b,
    ts_force nondet = unlifted_a unlifted_b lifted_a lifted_b
    ] "heap_lift_force_prevent.c"

context heap_lift_force_prevent begin

lemma heap_w32_hrs_mem [simp]:
    "heap_w32 (lift_global_heap s) p = Some a \<Longrightarrow> h_val (hrs_mem (t_hrs_' s)) p = a"
  by (metis h_val_simple_lift heap_lift_force_prevent.new_globals_ext_simps(3))

lemma lifted_a_wp [wp]:
    "\<lbrace> \<lambda>s. \<exists>a. heap_w32 s p = Some a \<and> P a s \<rbrace> lifted_a' p \<lbrace> \<lambda>r s. P r s \<rbrace>"
  by (clarsimp simp: lifted_a'_def, wp, auto)

lemma unlifted_a_wp [wp]:
    "\<lbrace> \<lambda>s. c_guard p \<and> P (h_val (hrs_mem (t_hrs_' s)) p) s \<rbrace>
                  unlifted_a' p \<lbrace> \<lambda>r s. P r s \<rbrace>"
  by (clarsimp simp: unlifted_a'_def, wp, auto)

lemma lifted_b_wp [wp]:
    "\<lbrace> \<lambda>s. \<exists>a. heap_w32 s p = Some a \<and> P (a * 3) s \<rbrace> lifted_b' p \<lbrace> \<lambda>r s. P r s \<rbrace>"
  apply (clarsimp simp: lifted_b'_def)
  apply wp
  apply (auto simp: simple_lift_c_guard lift_global_heap_def field_simps)
  done

lemma unlifted_b_wp [wp]:
    "\<lbrace> \<lambda>s. heap_ptr_valid (hrs_htd (t_hrs_' s)) p
           \<and> (\<forall>t. lift_global_heap t = lift_global_heap s \<longrightarrow> P (h_val (hrs_mem (t_hrs_' t)) p * 3) t) \<rbrace>
              unlifted_b' p \<lbrace> \<lambda>r s. P r s \<rbrace>"
  apply (clarsimp simp: unlifted_b'_def)
  apply wp
  apply (rule conjI)
   apply (metis simple_lift_c_guard simple_lift_def)
  apply clarsimp
  apply (rule_tac x="(h_val (hrs_mem (t_hrs_' s)) p)" in exI)
  apply (rule conjI)
   apply (clarsimp simp: lift_global_heap_def )
   apply (unfold simple_lift_def)
   apply (clarsimp  split: option.splits)
  apply (clarsimp simp: field_simps)
  apply (erule allE, erule (1) impE)
  apply (subgoal_tac "h_val (hrs_mem (t_hrs_' s)) p = h_val (hrs_mem (t_hrs_' t)) p")
   apply clarsimp
  apply (clarsimp simp: lift_global_heap_def)
  apply (drule fun_cong [where x=p])
  apply (unfold simple_lift_def [abs_def])
  apply (clarsimp split: split_if_asm)
  done

end

end

