(*
 * Copyright (C) 2013 NICTA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 *
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 *)

(* License: BSD, terms see file ./LICENSE *)

theory SepTactic
imports SepInv
begin

section "sep_point_tac"

definition sep_conj_extract :: "heap_assert \<Rightarrow> heap_assert" where
  "sep_conj_extract \<equiv> id"

definition sep_points :: "heap_assert \<Rightarrow> heap_assert" where
  "sep_points P \<equiv> P \<and>\<^sup>* sep_true"

lemma sep_conj_extract1D:
  "(P \<and>\<^sup>* Q) s \<Longrightarrow> sep_conj_extract (P \<and>\<^sup>* Q) s"
  by (simp add: sep_conj_extract_def)

lemma sep_conj_extract2D:
  "sep_conj_extract P s \<Longrightarrow> P s \<and> (sep_conj_extract P \<and>\<^sup>* sep_true) s"
apply(simp add: sep_conj_extract_def)
apply(erule sep_conj_sep_true)
done

lemma sep_conj_extract_assoc:
  "sep_conj_extract ((P \<and>\<^sup>* Q) \<and>\<^sup>* R) = sep_conj_extract (P \<and>\<^sup>* (Q \<and>\<^sup>* R))"
  by (simp add: sep_conj_ac)

lemma sep_conj_extract_decomposeD:
  "(sep_conj_extract (P \<and>\<^sup>* Q) \<and>\<^sup>* sep_true) s \<Longrightarrow> sep_points P s \<and>
      (sep_conj_extract Q \<and>\<^sup>* sep_true) s"
apply rule
 apply(simp add: sep_conj_extract_def sep_points_def sep_conj_ac)
 apply(erule (1) sep_conj_impl, simp)
apply(simp add: sep_conj_extract_def sep_conj_ac)
apply(subst (asm) sep_conj_assoc [symmetric])
apply(subst (asm) sep_conj_com)
apply(subst (asm) sep_conj_assoc)
apply(erule (1) sep_conj_impl)
apply simp
done

lemma sep_conj_extract_decomposeD2:
  "(sep_conj_extract P \<and>\<^sup>* sep_true) s \<Longrightarrow> sep_points P s"
  by (simp add: sep_conj_extract_def sep_points_def)

lemma sep_point_mapD:
  "sep_points (p \<mapsto>\<^sup>i\<^sub>g v) s \<Longrightarrow> (p \<hookrightarrow>\<^sup>i\<^sub>g v) s"
  by (simp add: sep_points_def sep_map_inv_def sep_map'_inv_def sep_map'_def
                sep_conj_ac)

lemma sep_point_map_excD:
  "sep_points (p \<mapsto>\<^sub>g v) s \<Longrightarrow> (p \<hookrightarrow>\<^sub>g v) s"
  by (simp add: sep_points_def sep_map'_def)

lemma sep_point_otherD:
  "sep_points P s \<Longrightarrow> True"
  by simp

ML {*

val sep_point_ds = [@{thm sep_point_mapD}, @{thm sep_point_map_excD}]

fun sep_point_tacs ds ctxt  = [
    REPEAT1 (dresolve_tac [@{thm sep_conj_extract1D}] 1),
    REPEAT1 (dresolve_tac [@{thm sep_conj_extract2D}] 1),
    clarify_tac ctxt 1,
    TRY (full_simp_tac (HOL_ss addsimps [@{thm sep_conj_extract_assoc}]) 1),
    REPEAT (dresolve_tac [@{thm sep_conj_extract_decomposeD},
        @{thm sep_conj_extract_decomposeD2}] 1 THEN
        clarify_tac ctxt 1),
    TRY (full_simp_tac (simpset_of ctxt) 1),
    REPEAT (dresolve_tac (sep_point_ds@ds) 1),
    REPEAT (dresolve_tac [@{thm sep_point_otherD}] 1),
    TRY (full_simp_tac (simpset_of ctxt) 1)
  ]

fun sep_point_tac ds ctxt = EVERY (sep_point_tacs ds ctxt)
*}

method_setup sep_point_tac =
  "Attrib.thms >> (fn thms => Method.SIMPLE_METHOD o sep_point_tac thms)"
  "extract points-to facts from separation assertions in assumptions"

lemma "(P \<and>\<^sup>* p \<mapsto>\<^sub>g v \<and>\<^sup>* Q) s \<Longrightarrow> (p \<hookrightarrow>\<^bsub>g\<^esub> v) s"
  by sep_point_tac


section "sep_exists_tac"

ML{*
val sep_exists_tac = full_simp_tac
    (HOL_ss addsimps [@{thm "sep_conj_exists1"}, @{thm "sep_conj_exists2"}])
*}

method_setup sep_exists_tac =
  "Scan.succeed (Method.SIMPLE_METHOD' o K sep_exists_tac)"
  "push existentials inside separation assertions to the outside"

lemma "(P \<and>\<^sup>* (\<lambda>s. (\<exists>x. Q x s))) s \<Longrightarrow> \<exists>x. (P \<and>\<^sup>* Q x) s"
  by sep_exists_tac

section "sep_select_tac"

definition sep_mark :: "heap_assert \<Rightarrow> heap_assert" where
  "sep_mark \<equiv> id"

definition sep_mark2 :: "heap_assert \<Rightarrow> heap_assert" where
  "sep_mark2 \<equiv> id"

lemma sep_mark2_id:
  "sep_mark2 P = P"
  by (simp add: sep_mark2_def)

lemma sep_markI:
  "(\<box> \<and>\<^sup>* sep_mark (P \<and>\<^sup>* Q)) s \<Longrightarrow> (P \<and>\<^sup>* Q) s"
  by (simp add: sep_mark_def sep_conj_empty)

lemma sep_mark_match:
  "(R \<and>\<^sup>* sep_mark2 P \<and>\<^sup>* Q) s \<Longrightarrow> (R \<and>\<^sup>* sep_mark (P \<and>\<^sup>* Q)) s"
  by (simp add: sep_mark_def sep_mark2_def)

lemma sep_mark_match2:
  "(R \<and>\<^sup>* sep_mark2 P) s \<Longrightarrow> (R \<and>\<^sup>* sep_mark P) s"
  by (simp add: sep_mark_def sep_mark2_def)

lemma sep_mark_mismatch:
  "((R \<and>\<^sup>* P) \<and>\<^sup>* sep_mark Q) s \<Longrightarrow> (R \<and>\<^sup>* sep_mark (P \<and>\<^sup>* Q)) s"
  by (simp add: sep_mark_def sep_conj_ac)

lemma sep_mark_mismatch2:
  "(R \<and>\<^sup>* P) s \<Longrightarrow> (R \<and>\<^sup>* sep_mark P) s"
  by (simp add: sep_mark_def)

lemma sep_emp_rem:
  "P s \<Longrightarrow> (\<box> \<and>\<^sup>* P) s"
  by (simp add: sep_conj_empty)

lemma sep_mark2_left:
  "(P \<and>\<^sup>* (sep_mark2 Q \<and>\<^sup>* R)) = (sep_mark2 Q \<and>\<^sup>* (P \<and>\<^sup>* R))"
  by (simp add: sep_mark2_def sep_conj_ac)

lemma sep_mark2_left2:
  "(P \<and>\<^sup>* sep_mark2 Q) = (sep_mark2 Q \<and>\<^sup>* P)"
  by (simp add: sep_mark2_def sep_conj_ac)

ML {*

(* Replace all occurences of an underscore that does not have an alphanumeric
   on either side of it *)
local
  val dummy_char = #" ";
  fun get_next (x::xs) = x
    | get_next [] = dummy_char;
  fun nth_id n = "?SEPTAC"^(Int.toString n);

  fun ok_char c = not (Char.isAlphaNum c);

  fun can_replace (prev,cur,last) =
        (ok_char prev) andalso (cur = #"_") andalso (ok_char last);

  fun replace_usc n prev_char (x::xs) =
        if can_replace (prev_char,x,get_next xs)
        then (nth_id n)::(replace_usc (n+1) x xs)
        else (String.str x)::(replace_usc (n+1) x xs)
    | replace_usc _ _ [] = []
in
  val rusc = String.concat o replace_usc 0 dummy_char o String.explode
end;

(* some tests *)
rusc "(_ +\<^isub>p _) \<mapsto> _";
rusc "_ \<mapsto>\<^isub>_ _";
rusc "_";
rusc "_ O_o _"
*}

ML{*

fun sep_select_tacs s ctxt  = [
    TRY (simp_tac (HOL_ss addsimps [@{thm sep_conj_assoc}]) 1),
    resolve_tac [@{thm sep_markI}] 1,
    REPEAT (res_inst_tac ctxt [(Lexicon.read_indexname "P",rusc s)] @{thm sep_mark_match} 1 ORELSE
      res_inst_tac ctxt [(Lexicon.read_indexname "P",rusc s)] @{thm sep_mark_match2} 1 ORELSE
      resolve_tac [@{thm sep_mark_mismatch}] 1 ORELSE
      resolve_tac [@{thm sep_mark_mismatch2}] 1),
    TRY (simp_tac (HOL_ss addsimps [@{thm sep_conj_assoc}]) 1),
    resolve_tac [@{thm sep_emp_rem}] 1,
    TRY (simp_tac (HOL_ss addsimps [@{thm sep_mark2_left},
      @{thm sep_mark2_left2}]) 1),
    TRY (simp_tac (HOL_ss addsimps [@{thm sep_mark2_id}]) 1)
  ]

fun sep_select_tac s ctxt = SELECT_GOAL (EVERY (sep_select_tacs s ctxt))

fun lift_parser p (ctxt,ts) =
  let val (r, ts') = p ts
  in (r, (ctxt,ts')) end
*}

method_setup sep_select_tac =
  {* lift_parser Args.name >> (fn s => Method.SIMPLE_METHOD' o sep_select_tac s) *}
  "takes a target conjunct in the goal and reorders it to the front"

(* Method.goal_args Args.name *)

lemma
  "\<And>R x f n. ((P::heap_assert) \<and>\<^sup>* fac x n \<and>\<^sup>* R (f x)) s"
apply(sep_select_tac "fac _ _")
apply(sep_select_tac "R _")
apply(sep_select_tac "fac x _")
apply(sep_select_tac "R _")
oops

lemma
  "((P::heap_assert) \<and>\<^sup>* fac x n) s"
apply(sep_select_tac "fac _ _")
oops

lemma
  "((P::heap_assert) \<and>\<^sup>* long_name) s"
  apply(sep_select_tac "long_name")
oops

consts c_guard :: "'a::c_type ptr_guard"

lemma sep_heap_update'_hrs:
  "(c_guard \<turnstile>\<^sub>s\<^sup>i p \<and>\<^sup>* (p \<mapsto>\<^sup>i\<^sub>c_guard v \<longrightarrow>\<^sup>* P)) (lift_state s) \<Longrightarrow>
      P (lift_state (hrs_mem_update (heap_update p (v::'a::mem_type)) s))"
  by (case_tac s)
     (simp add: hrs_mem_update_def, erule sep_heap_update')

lemma sep_map_lift_wp_hrs:
  "\<lbrakk> \<exists>v. (p \<mapsto>\<^sup>i\<^sub>c_guard v \<and>\<^sup>* (p \<mapsto>\<^sup>i\<^sub>c_guard v \<longrightarrow>\<^sup>* P v)) (lift_state s) \<rbrakk>
      \<Longrightarrow> P (lift (hrs_mem s) (p::'a::mem_type ptr)) (lift_state s)"
  by (case_tac s)
     (simp add: hrs_mem_def, erule sep_map_lift_wp)

lemma ptr_retyp_sep_cut'_wp_hrs:
  "\<lbrakk> (sep_cut' (ptr_val p) (size_of TYPE('a)) \<and>\<^sup>* (c_guard \<turnstile>\<^sub>s\<^sup>i p \<longrightarrow>\<^sup>* P))
      (lift_state s); c_guard (p::'a::mem_type ptr) \<rbrakk>
      \<Longrightarrow> P (lift_state (hrs_htd_update (ptr_retyp p) s))"
  by (case_tac s)
     (simp add: hrs_htd_update_def, erule (1) ptr_retyp_sep_cut'_wp)

ML {*

fun destruct bs (Bound n)      = Free (List.nth (bs,n))
  | destruct bs (Abs (s,ty,t)) = destruct ((s,ty)::bs) t
  | destruct bs (l$r)          = destruct bs l $ destruct bs r
  | destruct bs x              = x

fun schems (Abs (_,_,t)) = schems t
  | schems (l$r)         = schems l + schems r
  | schems (Var _)       = 1
  | schems _             = 0

fun hrs_mem_update (Const ("HeapRawState.hrs_mem_update",_)$_$s) = SOME s
  | hrs_mem_update _                                             = NONE

fun hrs_htd_update (Const ("HeapRawState.hrs_htd_update",_)$_$s) = SOME s
  | hrs_htd_update _                                             = NONE

fun lift_state_arg thy nc (Const ("TypHeap.lift_state",_)$s) = SOME s
  | lift_state_arg thy nc (l$r)                              =
        (case lift_state_arg thy nc r of
          SOME s => if nc orelse hrs_mem_update s <> NONE orelse
                        hrs_htd_update s <> NONE then SOME s
                        else lift_state_arg thy nc l
         | NONE   => lift_state_arg thy nc l)
  | lift_state_arg thy nc (Abs (_,_,t))                      =
        lift_state_arg thy nc t
  | lift_state_arg thy nc s =
        NONE (*error (Sign.string_of_term thy s)*)

fun term_of_thm thm = term_of (cprop_of thm)

fun lift_tac ctxt s old_schems =
  res_inst_tac ctxt [(Lexicon.read_indexname "s",s)] @{thm sep_map_lift_wp_hrs} 1 THEN
  (fn thm => if schems (term_of_thm thm) = old_schems then all_tac thm
                 else no_tac thm)

fun heap_update_tac ctxt s =
  res_inst_tac ctxt [(Lexicon.read_indexname "s",s)] @{thm sep_heap_update'_hrs} 1

fun ptr_retyp_tac ctxt s =
  res_inst_tac ctxt [(Lexicon.read_indexname "s",s)] @{thm ptr_retyp_sep_cut'_wp_hrs} 1

fun sep_wp_step_tac' ctxt lift_only s old_schems = let
  fun prt s = (
    case s of
        Free (a, b) => a
      | _ => raise TERM ("Not a free variable",[s]))
  val state = prt s
  val step = case hrs_mem_update s of
               SOME s => heap_update_tac ctxt (prt s)
             | NONE => case hrs_htd_update s of
                         SOME s => ptr_retyp_tac ctxt (prt s)
                       | NONE => no_tac
in
  TRY (lift_tac ctxt state old_schems) THEN (if lift_only then all_tac else step)
end

fun sep_wp_step_tac ctxt lift_only thm = let
  val t = term_of_thm thm
  val thy = Proof_Context.theory_of ctxt
  val old_schems = schems t
  val s = lift_state_arg thy lift_only (destruct [] t)
in
  (case s of
    SOME state => sep_wp_step_tac' ctxt lift_only state old_schems
   | NONE => no_tac) thm
end

fun sep_wp_tac ctxt = (REPEAT (sep_wp_step_tac ctxt false)) THEN sep_wp_step_tac ctxt true

*}

method_setup sep_wp_tac =
  "Scan.succeed (Method.SIMPLE_METHOD o sep_wp_tac)"
  "apply WP separation logic rules to goal"


(* see testfiles/sep_example_pre_list.thy for a more detailed example *)
lemma
  "((\<lambda>z. lift (hrs_mem s) (p::(32 word) ptr) + 1 = 2) \<and>\<^sup>* P) (lift_state s)"
apply sep_wp_tac
oops


end
